/*
  Program: Hra Zajc a psi
  Autor: Radek Pelnek  
  Ploha ke knize "Jak to vyeit",  www.radekpelanek.cz/?jtv

  Zadn: Napite program, kter bude hrt hru "Zajc a psi" proti lovku pro
  hrac pln velikosti 6 x 6 (tj. 3 psi). Program nech hre zvolit,
  zda chce hrt za zajce nebo za psi, pot program opakovan vdy vype
  aktuln stav hracho plnu v jednoduchm textovm ztvrnn, zept se na tah
  hre, tah provede, vybere svj tah a ten provede. Program by ml hrt
  optimln strategii, tj. pokud dostane pleitost, vyhraje.

  Poznmka: Tento program je napsn vcelku "nzkorovovn" a slou spe jako
  ukzka, jak lze problm vyeit pomoc relativn krtkho kdu (bez koment
  cca 180 dk, z toho asi 110 je vlastn jdro een, zbytek je "obsluha").
  Program tak ilustruje nkter "triky", kdyby ml bt kd soust vtho
  projektu, bylo by lep pepsat jej teba objektov - pak by byl del, ale
  itelnj.
*/


#include <iostream>
#include <string>
#include <map>
#include <cstdlib>

#define N 3
#define ZACINA_ZAJIC false
// Pocet ps = velikost plnu, v principu funguje i pro vt hodnoty N, ale
// prakticky dobhne maximln pro 4.

using namespace std;

/*  Prohledvn je napsno tak, e po vystame pouze s jednm stavem - dlme
    prohledvn do hloubky a pi vracen zpt "odinme" krok dopedu. 
    Potebujeme si v kadou chvli tedy pamatovat pouze jeden stav - ten drme
    v ne uvedench globlnch strukturch.
*/
int pesx[N], pesy[N];  // poloha ps
int zajicx, zajicy;    // poloha zajce
bool zajic_na_tahu;    // kdo je na tahu

/* Ve struktue optimalni_tah si udrujeme pro kad stav, kdo v nm vtz a
   ppadn jakm tahem. Kd stavu udv ne popsan funkce, kd tahu je nsledujc:
    * 0-3 znaci pohyb  zajce (sud = vlevo, lich = vpravo) (0,1 nahoru, 2,3 dol)
    * vy sla pohyb ps  (sud = vlevo, lich = vpravo)
 */
map<string, int> optimalni_tah;

void nastav_inicialni_stav() {
  for (int i=0; i<N; i++) {
    pesx[i] = 2*i + 1;
    pesy[i] = 0;
  }
  zajicx = 0; zajicy = 2*N-1;
  zajic_na_tahu = ZACINA_ZAJIC;
}

/* Vrt textov kd aktulnho stavu. Potebujeme pouze, aby kd byl
   jednoznan (bez koliz), jinak je jedno, jak bude vypadat. Rozliujeme
   stavy podle poad ps - to je teoreticky redundantn a vede to k zbytenmu
   nrstu potu prohledanch stav. Prakticky to m tu vhodu, e se nm
   jednodueji reprezentuj vtzn tahy. Pro N=3 to nen zsadn, pro vy N
   by bylo poteba pedlat kdovn rozumnji.
 */

string kod_aktualniho_stavu() {
  string kod = zajic_na_tahu ? "z":"p";
  for (int j=0; j<2*N; j++)
    for (int i=0; i<2*N; i++)
      if ((i+j)%2 == 1) {
	bool prazdne_pole = true;
	for (int k=0; k<N;k++)
	  if (pesx[k]==i && pesy[k]==j) {
	    kod += '0' + (k + 1);
	    prazdne_pole = false;
	  } 
	if (zajicx==i && zajicy==j) {kod+="X"; prazdne_pole = false;}
	if (prazdne_pole) kod += ".";
      }
  return kod;
}

void vypis_stav() {
  int i,j,k;

  cout << endl << "Na tahu: "<< (zajic_na_tahu?"zajic":"psi")<<endl;
  // odkomentovnm nsledujcho dku dostaneme vpis optimlnch tah v prbhu hry
  //  cout <<  "Optimalni tah: "<< (zajic_na_tahu? optimalni_tah[kod_aktualniho_stavu()]: optimalni_tah[kod_aktualniho_stavu()]-4)<<endl;
  for (i=0; i<2*N+2; i++) cout << "#";
  cout <<endl;
  int pes = 1;
  for (j=0; j<2*N; j++) {
    cout << "#";
    for (i=0; i<2*N; i++) {
      if ((i+j) % 2 == 0) { cout << " "; }
      else {
 	bool empty = true;
 	for (k=0; k<N; k++)
 	  if (pesx[k] == i && pesy[k] == j) { cout << k+1; empty = false; }
 	if (zajicx == i && zajicy ==j ) { cout << "X"; empty = false;}
 	if (empty) cout << ".";	
      }
    }
    cout << "#"<< endl;
  }
  for (i=0; i<2*N+2; i++) cout << "#";
  cout <<endl;
}

/* Zmn aktuln stav podle uvedenho tahu (kdovn viz nahoe), vetn
 kontroly platnosti tahu. Odpovd pravidlm hry. */
bool udelej_tah(int tah) {
  bool platny = true;
  int novex, novey;
  // pro uivatele a prohledvn kdujeme tahy od 1, ale pro vnitn vyhodnocen od 0
  // v krtkm programu to m svj vznam, v jakmkoli vtm projektu by to ovem
  // nebylo rozumn, protoe by to tm jist vedlo ke zmatkm 
  tah--;   
  
  if (tah < 4) { // zajc
    if (!zajic_na_tahu) platny = false;
    novex = zajicx; novey = zajicy;
    if (tah % 2 == 0) novex--; else novex++;
    if (tah < 2) novey--; else novey++;
    if (novex < 0 || novey<0 || 2*N <=novex || 2*N <= novey) platny = false;
    for (int k=0; k<N;k++) 
      if (pesx[k] == novex && pesy[k] == novey) platny = false;        
    if (platny) { zajicx = novex; zajicy = novey;}
  } else { // psi
    if (zajic_na_tahu) platny = false;
    tah -= 4;
    novex = pesx[tah/2]; novey = pesy[tah/2];
    novey++;
    if (tah%2 ==0) novex--; else novex++;
    if (novex<0 || novey<0 || 2*N <=novex || 2*N <= novey) platny = false;
    for (int k=0; k<N;k++) 
      if (k!= tah/2 && (pesx[k] ==novex && pesy[k] ==novey)) platny = false;
    if (novex==zajicx && novey == zajicy) platny = false;
    if (platny) { pesx[tah/2] = novex; pesy[tah/2]=novey;}	  
  }
  if (platny) { zajic_na_tahu = ! zajic_na_tahu; }
  return platny;
}

/*  Udl reverzn tah. Pedpokld, e reverze je korektn a nedl kontroly. */
void vrat_tah(int tah) {
  tah--; 
  if (tah < 4) { //zajc
    if (tah % 2 == 0) zajicx++; else zajicx--;
    if (tah < 2) zajicy++; else zajicy--;    
  } else { // psi
    tah -=4;
    pesy[tah/2]--;
    if (tah % 2 == 0) pesx[tah/2]++; else pesx[tah/2]--;
  }
  zajic_na_tahu = ! zajic_na_tahu;
}

/* Rekurzivn prohledn vech stav pomoc DFS a pitom souasn buduje
   ohodnocen (minimax algoritmus).

   Aktuln stav je uloen v globlnch promnnch.
 */

void prohledavani() {
  int opt_tah = -1;

  // vyzkouej vechny mon tahy, tady sluji tahy od 1, abych odliil prvn tah od nedefinovanho tahu
  for (int tah=1; tah <= 4+N*2; tah++) {
    if (udelej_tah(tah)) {
      // byl to platn tah - jsem v nslednkovi
      // pokud nslednk jet nebyl vyhodnocen, rekurzivn se zanome
      if (! optimalni_tah[kod_aktualniho_stavu()]) {
	prohledavani();
      }
      // te u nsledujc stav mus bt ohodnocen
      // pokud nasledujc stav nem optimalni tah, tak ho m aktuln stav - thnu do tohoto nslednka
      if (optimalni_tah[kod_aktualniho_stavu()] == -1) {
	opt_tah = tah;
      }
      // udlme reverzn tah a vrtme se do aktulnho stavu
      vrat_tah(tah);
    }
  }
  optimalni_tah[kod_aktualniho_stavu()] = opt_tah;
  // vpis 
  cout <<kod_aktualniho_stavu()<< " -> " << opt_tah << endl;
}

/* Udl nhodn platn tah
 */
bool nahodny_tah() {
  int platne_tahy = 0;
  int tah;
  // spotm poet platnch tah
  for (tah=1; tah <= 4+N*2; tah++) {
    if (udelej_tah(tah)) {
      platne_tahy++;
      vrat_tah(tah);
    }
  }
  if (platne_tahy==0) return false;
  // vyberu nhodn z nich a provedu ho
  int r = rand() % platne_tahy;
  for (tah=1; tah <= 4+N*2; tah++) {
    if (udelej_tah(tah)) {
      if (r==0) return true;
      r--;
      vrat_tah(tah);
    }
  }
  return false;
}

// nate tah od hre
bool tah_hrace() {  
  if (zajic_na_tahu) {
    cout << "1 = nahoru vlevo\n2 = nahoru vpravo\n3 = dolu vlevo\n4 = dolu vpravo\n";
  } else {
    for (int k=0; k<N;k++) {
      cout << k*2+1 << " = " << k+1 <<"-ty vlevo"<<endl; 
      cout << k*2+2 << " = " << k+1 <<"-ty vpravo"<<endl; 
    }
  }
  cout <<"0 = konec"<<endl;
  int tah;
  bool done = false;
  // nata dokud nedostane platn tah
  while (!done) {
    cin >> tah;
    if (tah == 0) return false;
    if (!zajic_na_tahu) {tah +=4;}
    if (tah > 0 && tah <= 2*N+4)
      done = udelej_tah(tah);
  }
  return true;
}

// udl tah potae - pokud lze tak optimln, jinak nhodn
bool tah_pocitace() {
  int tah = optimalni_tah[kod_aktualniho_stavu()];
  if (tah != -1) {
    udelej_tah(tah);
  } else {
    if (!nahodny_tah()) return false;
  }
  return true;
}

// sehraje hru potae proti lovku
void hraj() {
  bool hrac_na_tahu = true;
  char c;
  bool konec = false;
  
  cout << "Chces zacinat? [A/N]"<<endl;
  cin>> c;
  hrac_na_tahu = (c=='A' || c=='a');

  while (!konec) {
    vypis_stav();
    if (hrac_na_tahu) {
      konec = ! tah_hrace();
    } else {
      konec = ! tah_pocitace();
    }
    hrac_na_tahu = ! hrac_na_tahu;
  }    
}

int main() {

  srand(time(NULL));
  
  nastav_inicialni_stav();
    
  prohledavani();
  
  cout << "Pocet prohledanych stavu: " << optimalni_tah.size() << endl;

  hraj();
  
  return 0;
}
